/*
 * aaccess.cpp
 *
 *  Created on: 26. 3. 2014
 *	  Author: Jan Travnicek
 */

#include <tclap/CmdLine.h>
#include <global/GlobalData.h>
#include <measure>

#include <exception/CommonException.h>
#include <lexer/Lexer.h>
#include <parser/Parser.h>

#include <factory/XmlDataFactory.hpp>

#include "access/AutomatonAccess.h"
#include "access/GrammarAccess.h"
#include "access/RegExpAccess.h"
#include "access/StringAccess.h"
#include "access/TreeAccess.h"

#include "access/settings/AutomatonSettings.h"
#include "access/settings/GrammarSettings.h"
#include "access/settings/RegExpSettings.h"
#include "access/settings/StringSettings.h"
#include "access/settings/TreeSettings.h"

int main ( int argc, char * argv[] ) {
	try {
		common::GlobalData::argc = argc;
		common::GlobalData::argv = argv;

		TCLAP::CmdLine cmd ( "Access internals of automata, grammars, regexps, ...", ' ', "0.01" );

		// ----------------------------------------------------------------------------------------------------------------------------------------------------------------------
		std::vector < std::string > automatonSettings {
			dataAccess::AutomatonSettings::stringsVector ( )
		};

		TCLAP::ValuesConstraint < std::string > automatonPrintingOptionsVals ( automatonSettings );

		TCLAP::ValueArg < std::string > automatonInput ( "", "automaton", "Access components of automata", false, "", & automatonPrintingOptionsVals );

		// ----------------------------------------------------------------------------------------------------------------------------------------------------------------------

		std::vector < std::string > grammarSettings {
			dataAccess::GrammarSettings::stringsVector ( )
		};

		TCLAP::ValuesConstraint < std::string > grammarPrintingOptionsVals ( grammarSettings );

		TCLAP::ValueArg < std::string > grammarInput ( "", "grammar", "Access components of grammars", false, "", & grammarPrintingOptionsVals );

		// ----------------------------------------------------------------------------------------------------------------------------------------------------------------------

		std::vector < std::string > regexpSettings {
			dataAccess::RegExpSettings::stringsVector ( )
		};

		TCLAP::ValuesConstraint < std::string > regexpPrintingOptionsVals ( regexpSettings );

		TCLAP::ValueArg < std::string > regexpInput ( "", "regexp", "Access components of regexps", false, "", & regexpPrintingOptionsVals );

		// ----------------------------------------------------------------------------------------------------------------------------------------------------------------------

		std::vector < std::string > stringSettings {
			dataAccess::StringSettings::stringsVector ( )
		};

		TCLAP::ValuesConstraint < std::string > stringPrintingOptionsVals ( stringSettings );

		TCLAP::ValueArg < std::string > stringInput ( "", "string", "Access components of strings", false, "", & stringPrintingOptionsVals );

		// ----------------------------------------------------------------------------------------------------------------------------------------------------------------------

		std::vector < std::string > treeSettings {
			dataAccess::TreeSettings::stringsVector ( )
		};

		TCLAP::ValuesConstraint < std::string > treePrintingOptionsVals ( treeSettings );

		TCLAP::ValueArg < std::string > treeInput ( "", "tree", "Access components of trees", false, "", & treePrintingOptionsVals );

		// ----------------------------------------------------------------------------------------------------------------------------------------------------------------------

		std::vector < TCLAP::Arg * > xorlist;
		xorlist.push_back ( & automatonInput );
		xorlist.push_back ( & grammarInput );
		xorlist.push_back ( & regexpInput );
		xorlist.push_back ( & stringInput );
		xorlist.push_back ( & treeInput );
		cmd.xorAdd ( xorlist );

		std::vector < std::string > operationSettings {
			"get", "set", "add", "remove"
		};
		TCLAP::ValuesConstraint < std::string > allowedVals ( operationSettings );

		TCLAP::ValueArg < std::string > operation ( "o", "operation", "Operation type", false, "", & allowedVals );
		cmd.add ( operation );

		TCLAP::ValueArg < std::string > file ( "i", "input", "Read from file", false, "-", "file" );
		cmd.add ( file );

		TCLAP::ValueArg < std::string > element ( "e", "element", "Operation argument", false, "-", "file" );
		cmd.add ( element );

		TCLAP::ValueArg < std::string > collection ( "c", "collection", "Operation arguments collection", false, "-", "file" );
		cmd.add ( collection );

		TCLAP::SwitchArg measure ( "m", "measure", "Measure times", false );
		cmd.add ( measure );

		TCLAP::SwitchArg verbose ( "v", "verbose", "Be verbose", false );
		cmd.add ( verbose );

		cmd.parse ( argc, argv );

		if ( verbose.isSet ( ) )
			common::GlobalData::verbose = true;

		if ( measure.isSet ( ) )
			common::GlobalData::measure = true;

		cli::Environment environment;
		environment.setBinding ( "stdin", file.getValue ( ) );
		environment.setBinding ( "element", element.getValue ( ) );
		environment.setBinding ( "collection", collection.getValue ( ) );
		environment.setBinding ( "stdout", "-" );

		measurements::start ( "Overal", measurements::Type::OVERALL );
		measurements::start ( "Input read", measurements::Type::AUXILIARY );

		cli::Parser parser ( cli::Lexer ( "execute <#stdin > $data" ) );
		parser.parse ( )->run ( environment );

		if ( collection.isSet ( )
		&& ( operation.getValue ( ) == "set" || operation.getValue ( ) == "add" || operation.getValue ( ) == "remove" )
		&& ( automatonInput.isSet ( ) || grammarInput.isSet ( ) || regexpInput.isSet ( ) || stringInput.isSet ( ) ) ) {
			parser = cli::Parser ( cli::Lexer ( "execute < :set @Object #collection > $argument" ) );
			parser.parse ( )->run ( environment );
		}
		if ( collection.isSet ( )
		&& ( operation.getValue ( ) == "add" || operation.getValue ( ) == "remove" ) && ( treeInput.isSet ( ) ) ) {
			parser = cli::Parser ( cli::Lexer ( "execute < :set @ranked_symbol #collection > $argument" ) );
			parser.parse ( )->run ( environment );
		}

		if ( element.isSet ( )
		&& ( operation.getValue ( ) == "set" || operation.getValue ( ) == "add" || operation.getValue ( ) == "remove" )
		&& ( automatonInput.isSet ( ) || grammarInput.isSet ( ) || regexpInput.isSet ( ) || stringInput.isSet ( ) ) ) {
			parser = cli::Parser ( cli::Lexer ( "execute < :Object #element > $argument" ) );
			parser.parse ( )->run ( environment );
		}
		if ( element.isSet ( ) && ( operation.getValue ( ) == "add" || operation.getValue ( ) == "remove" ) && ( treeInput.isSet ( ) ) ) {
			parser = cli::Parser ( cli::Lexer ( "execute < :ranked_symbol #element > $argument" ) );
			parser.parse ( )->run ( environment );
		}

		measurements::end();
		measurements::start ( "Algorithm", measurements::Type::MAIN );

		std::string cliString;
		if ( automatonInput.isSet ( ) ) {
			cliString = "execute " + dataAccess::AutomatonAccess::access ( automatonInput.getValue ( ) ) + "::" + operation.getValue ( ) + " $data";
		} else if ( grammarInput.isSet ( ) ) {
			cliString = "execute " + dataAccess::GrammarAccess::access ( grammarInput.getValue ( ) ) + "::" + operation.getValue ( ) + " $data";
		} else if ( regexpInput.isSet ( ) ) {
			cliString = "execute " + dataAccess::RegExpAccess::access ( regexpInput.getValue ( ) ) + "::" + operation.getValue ( ) + " $data";
		} else if ( stringInput.isSet ( ) ) {
			cliString = "execute " + dataAccess::StringAccess::access ( stringInput.getValue ( ) ) + "::" + operation.getValue ( ) + " $data";
		} else if ( treeInput.isSet ( ) ) {
			cliString = "execute " + dataAccess::TreeAccess::access ( treeInput.getValue ( ) ) + "::" + operation.getValue ( ) + " $data";
		} else {
			throw exception::CommonException ( "Input not recognized." );
		}

		if ( operation.getValue ( ) == "add" || operation.getValue ( ) == "remove" )
			cliString += " $argument";

		if ( operation.getValue ( ) == "get" ) {
			cliString += " >$output";
		}

		parser = cli::Parser ( cli::Lexer ( cliString ) );
		parser.parse ( )->run ( environment );

		measurements::end();
		measurements::start("Output write", measurements::Type::AUXILIARY);

		if ( operation.getValue ( ) == "get" ) {
			parser = cli::Parser ( cli::Lexer ( "execute $output >#stdout" ) );
		} else {
			parser = cli::Parser ( cli::Lexer ( "execute $data >#stdout" ) );
		}
		parser.parse ( )->run ( environment );

		measurements::end();
		measurements::end();

		if ( measure.getValue ( ) ) common::Streams::measure << measurements::results ( ) << std::endl;

		return 0;
	} catch ( const exception::CommonException & exception ) {
		factory::XmlDataFactory::toStdout ( exception );
		return 1;
	} catch ( const TCLAP::ArgException & exception ) {
		common::Streams::out << exception.error ( ) << std::endl;
		return 2;
	} catch ( const std::exception & exception ) {
		common::Streams::err << "Exception caught: " << exception.what ( ) << std::endl;
		return 3;
	} catch ( ... ) {
		common::Streams::err << "Unknown exception caught." << std::endl;
		return 127;
	}
}
