/*
 * ConversionHandler.cpp
 *
 *  Created on: 23. 2. 2014
 *	  Author: Tomas Pecka
 */

#include "ConversionHandler.h"
#include <exception/CommonException.h>

ConversionHandler::ConversionHandler( const std::string & inputType, const std::string & target, const std::string & algorithm ) :
	m_source( parseFormalismFromType( inputType ) ),
	m_target( parseFormalismFromString( target ) ),
	m_algorithm( parseAlgorithmFromString( algorithm ) )
{
}

std::string ConversionHandler::convert ( ) {
	if ( m_source == FINITE_AUTOMATON )
		return convertFA ( );
	else if ( m_source == REGULAR_EXPRESSION )
		return convertRE ( );
	else if ( m_source == REGULAR_GRAMMAR )
		return convertRG ( );
	else if ( m_source == PUSHDOWN_AUTOMATON )
		return convertPDA ( );
	else if ( m_source == CONTEXTFREE_GRAMMAR )
		return convertCFG ( );
	else if ( m_source == REGULAR_TREE_EXPRESSION )
		return convertRTE ( );
	else
		throw exception::CommonException( "ConversionHandler: Unknown source formalism." );
}

// ----------------------------------------------------------------------------

std::string ConversionHandler::convertFA ( ) {
	if ( m_target == REGULAR_EXPRESSION )
		return convertFAtoRE ( );
	else if ( m_target == REGULAR_GRAMMAR )
		return convertFAtoRG ( );
	else
		throw exception::CommonException( "ConversionHandler: Unknown target formalism." );
}

std::string ConversionHandler::convertRE ( ) {
	if ( m_target == FINITE_AUTOMATON )
		return convertREtoFA ( );
	else if ( m_target == REGULAR_GRAMMAR )
		return convertREtoRG ( );
	else
		throw exception::CommonException( "ConversionHandler: Invalid target formalism." );
}

std::string ConversionHandler::convertRG ( ) {
	if ( m_target == FINITE_AUTOMATON )
		return convertRGtoFA ( );
	else if ( m_target == REGULAR_EXPRESSION )
		return convertRGtoRE ( );
	else
		throw exception::CommonException( "ConversionHandler: Invalid target formalism." );
}

std::string ConversionHandler::convertPDA ( ) {
	throw exception::CommonException ( "ConversionHandler: Invalid source formalism." );
}

std::string ConversionHandler::convertCFG ( ) {
	if ( m_target == PUSHDOWN_AUTOMATON )
		return convertCFGtoPDA ( );
	else
		throw exception::CommonException ( "ConversionHandler: Invalid target formalism." );
}

std::string ConversionHandler::convertRTE ( ) {
	if ( m_target == PUSHDOWN_AUTOMATON )
		return convertRTEtoPDA ( );
	else
		throw exception::CommonException( "ConversionHandler: Invalid target formalism." );
}

std::string ConversionHandler::convertFTA ( ) {
	if ( m_target == PUSHDOWN_AUTOMATON )
		return convertFTAtoPDA ( );
	else
		throw exception::CommonException( "ConversionHandler: Invalid target formalism." );
}

// ----------------------------------------------------------------------------

std::string ConversionHandler::convertFAtoRE ( ) {
	switch ( m_algorithm ) {
	case BRZOZOWSKI_ALGEBRAIC:
		return "execute automaton::convert::ToRegExpAlgebraic $input > $output";
	case STATE_ELIMINATION:
	case DEFAULT:
		return "execute automaton::convert::ToRegExpStateElimination $input > $output";
	default:
		throw exception::CommonException ( "ConversionHandler: Invalid algorithm." );
	}
}

std::string ConversionHandler::convertFAtoRG ( ) {
	switch ( m_algorithm ) {
	case INCOMING_TRANSITIONS:
		return "execute automaton::convert::ToGrammarLeftRG $input > $output";
	case OUTGOING_TRANSITIONS:
	case DEFAULT:
		return "execute automaton::convert::ToGrammarRightRG $input > $output";
	default:
		throw exception::CommonException( "ConversionHandler: Invalid algorithm." );
	}
}

// ----------------------------------------------------------------------------

std::string ConversionHandler::convertREtoFA ( ) {
	switch ( m_algorithm ) {
	case BRZOZOWSKI_DERIVATION:
		return "execute regexp::convert::ToAutomatonDerivation $input > $output";
	case THOMPSON_NFA:
		return "execute regexp::convert::ToAutomatonThompson $input > $output";
	case GLUSHKOV_NFA:
	case DEFAULT:
		return "execute regexp::convert::ToAutomatonGlushkov $input > $output";
	default:
		throw exception::CommonException( "ConversionHandler: Invalid algorithm." );
	}
}

std::string ConversionHandler::convertREtoRG ( ) {
	switch ( m_algorithm ) {
	case BRZOZOWSKI_DERIVATION:
		return "execute regexp::convert::ToGrammarRightRGDerivation $input > $output";
	case GLUSHKOV_NFA:
	case DEFAULT:
		return "execute regexp::convert::ToGrammarRightRGGlushkov $input > $output";
	default:
		throw exception::CommonException( "ConversionHandler: Invalid algorithm." );
	}
}

// ----------------------------------------------------------------------------

std::string ConversionHandler::convertRGtoFA( ) {
	switch ( m_algorithm ) {
	case DEFAULT:
		return "execute grammar::convert::ToAutomaton $input > $output";
	default:
		throw exception::CommonException( "ConversionHandler: Invalid algorithm." );
	}
}

std::string ConversionHandler::convertRGtoRE ( ) {
	switch ( m_algorithm ) {
	case BRZOZOWSKI_ALGEBRAIC:
	case DEFAULT:
		return "execute grammar::convert::ToRegExpAlgebraic $input > $output";
	default:
		throw exception::CommonException( "ConversionHandler: Invalid algorithm." );
	}
}

std::string ConversionHandler::convertCFGtoPDA ( ) {
	switch ( m_algorithm ) {
	case TOP_DOWN:
		return "execute grammar::convert::ToAutomaton $input > $output";
	case BOTTOM_UP:
		return "execute grammar::convert::ToAutomatonBottomUp $input > $output";
	default:
		throw exception::CommonException ( "ConversionHandler: Invalid algorithm." );
	}
}

std::string ConversionHandler::convertRTEtoPDA ( ) {
	switch ( m_algorithm ) {
	case GLUSHKOV_RTE_NAIVE:
		return "execute rte::convert::ToPostfixPushdownAutomatonGlushkovNaive $input > $output";
	case GLUSHKOV_RTE:
		return "execute rte::convert::ToPostfixPushdownAutomatonGlushkov $input > $output";
	default:
		throw exception::CommonException ( "ConversionHandler: Invalid algorithm." );
	}
}

std::string ConversionHandler::convertFTAtoPDA ( ) {
	switch ( m_algorithm ) {
	case DEFAULT:
		return "execute automaton::convert::ToPostfixPushdownAutomaton $input > $output";
	default:
		throw exception::CommonException( "ConversionHandler: Invalid algorithm." );
	}
}

// ----------------------------------------------------------------------------

ConversionHandler::TFormalism ConversionHandler::parseFormalismFromString ( const std::string & target ) const {
	if ( target == "fa" ) return FINITE_AUTOMATON;

	if ( target == "re" ) return REGULAR_EXPRESSION;

	if ( target == "rg" ) return REGULAR_GRAMMAR;

	if ( target == "pda" ) return PUSHDOWN_AUTOMATON;

	if ( target == "cfg" ) return CONTEXTFREE_GRAMMAR;

	if ( target == "rte" ) return REGULAR_TREE_EXPRESSION;

	throw exception::CommonException ( "ConversionHandler: Unknown target type." );
}

ConversionHandler::TFormalism ConversionHandler::parseFormalismFromType ( const std::string & type ) const {
	if ( type.find ( "automaton::" ) == 0 )
		return FINITE_AUTOMATON;

	if ( type.find ( "regexp::" ) == 0 )
		return REGULAR_EXPRESSION;

	if ( type.find ( "grammar::CFG" ) == 0 )
		return CONTEXTFREE_GRAMMAR;

	if ( type.find ( "grammar::" ) == 0 )
		return REGULAR_GRAMMAR;

	if ( type.find ( "rte::" ) == 0 )
		return REGULAR_TREE_EXPRESSION;

	throw exception::CommonException ( "ConversionHandler: Invalid input formalism." );
}

ConversionHandler::TAlgorithm ConversionHandler::parseAlgorithmFromString ( const std::string & algorithm ) const {
	if ( algorithm == "elimination" ) return STATE_ELIMINATION;

	if ( algorithm == "algebraic" ) return BRZOZOWSKI_ALGEBRAIC;

	if ( algorithm == "brzozowski" ) return BRZOZOWSKI_DERIVATION;

	if ( algorithm == "glushkov" ) return GLUSHKOV_NFA;

	if ( algorithm == "thompson" ) return THOMPSON_NFA;

	if ( algorithm == "incoming" ) return INCOMING_TRANSITIONS;

	if ( algorithm == "outgoing" ) return OUTGOING_TRANSITIONS;

	if ( algorithm == "topdown" ) return TOP_DOWN;

	if ( algorithm == "bottomup" ) return BOTTOM_UP;

	if ( algorithm == "glushkovrtenaive" ) return GLUSHKOV_RTE_NAIVE;

	if ( algorithm == "glushkovrte" ) return GLUSHKOV_RTE;

	if ( ( algorithm == "" ) || ( algorithm == "default" ) ) return DEFAULT;

	throw exception::CommonException( "ConversionHandler:: Invalid algorithm. See help." );
}
