/*
 * Pair.h
 *
 * This file is part of Algorithms library toolkit.
 * Copyright (C) 2017 Jan Travnicek (jan.travnicek@fit.cvut.cz)

 * Algorithms library toolkit is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Algorithms library toolkit is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Algorithms library toolkit.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Created on: Apr 1, 2013
 * Author: Jan Travnicek
 */

#ifndef OBJECTS_PAIR_H_
#define OBJECTS_PAIR_H_

#include <alib/pair>
#include <alib/string>
#include <sstream>

#include "ContainerBase.h"

#include <object/UniqueObject.h>
#include <object/ObjectFactory.h>

#include <core/normalize.hpp>

namespace container {

/**
 * \brief
 * Represents an adaptor of a pair datatype from the stl extensions provided by the algorithms library.
 *
 * The datatype packs two arbitrary values together.
 *
 * \tparam FirstType the type of the first object in the pair.
 * \tparam SecondType the type of the second object in the pair.
 */
template < class FirstType = object::Object, class SecondType = object::Object >
class ObjectsPair final : public ext::pair < FirstType, SecondType >, public ContainerBase {
public:
	/**
	 * \brief
	 * Creates a new instance of the pair based on the two packed values.
	 *
	 * \param firstObject the first value of the pair
	 * \param firstObject the second value of the pair
	 */
	explicit ObjectsPair ( FirstType firstObject, SecondType secondObject );

	/**
	 * \brief
	 * Creates a new instance of the pair based on the extended version of the stl pair.
	 *
	 * \param raw the extended pair datatype
	 */
	explicit ObjectsPair ( ext::pair < FirstType, SecondType > raw );

	/**
	 * @copydoc alphabet::ContainerBase::clone ( ) const &
	 */
	virtual ContainerBase * clone ( ) const & override;

	/**
	 * @copydoc alphabet::ContainerBase::clone ( ) &&
	 */
	virtual ContainerBase * clone ( ) && override;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::compare ( const ObjectBase & ) const
	 */
	virtual int compare ( const ObjectBase & other ) const override {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	/**
	 * The actual compare method
	 *
	 * \param other the other instance
	 *
	 * \returns the actual relation between two by type same containers
	 */
	int compare ( const ObjectsPair & other ) const;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::operator >> ( std::ostream & ) const
	 */
	virtual void operator >>( std::ostream & os ) const override;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::operator std::string ( ) const
	 */
	virtual explicit operator std::string ( ) const override;

	/**
	 * @copydoc object::ObjectBase::inc ( ) &&
	 */
	virtual object::ObjectBase * inc ( ) && override;
};

template < class FirstType, class SecondType >
ObjectsPair < FirstType, SecondType >::ObjectsPair(ext::pair<FirstType, SecondType> raw) : ext::pair<FirstType, SecondType>(std::move(raw)) {

}

template < class FirstType, class SecondType >
ObjectsPair < FirstType, SecondType >::ObjectsPair(FirstType firstObject, SecondType secondObject) : ext::pair<FirstType, SecondType>(std::move(firstObject), std::move(secondObject)) {

}

template < class FirstType, class SecondType >
ContainerBase* ObjectsPair < FirstType, SecondType >::clone ( ) const & {
	return new ObjectsPair(*this);
}

template < class FirstType, class SecondType >
ContainerBase* ObjectsPair < FirstType, SecondType >::clone() && {
	return new ObjectsPair(std::move(*this));
}

template < class FirstType, class SecondType >
int ObjectsPair < FirstType, SecondType >::compare(const ObjectsPair& other) const {
	static ext::compare<ext::pair<FirstType, SecondType>> comp;
	return comp ( static_cast < const ext::pair < FirstType, SecondType > & > ( * this ), static_cast < const ext::pair < FirstType, SecondType > & > ( other ) );
}

template < class FirstType, class SecondType >
void ObjectsPair < FirstType, SecondType >::operator>>(std::ostream& os) const {
	os << "(ObjectsPair " << static_cast < const ext::pair < FirstType, SecondType > & > ( * this ) << ")";
}

template < class FirstType, class SecondType >
ObjectsPair < FirstType, SecondType >::operator std::string() const {
	std::stringstream ss;
	ss << *this;
	return std::move(ss).str();
}

template < class FirstType, class SecondType >
object::ObjectBase* ObjectsPair < FirstType, SecondType >::inc() && {
	return new object::UniqueObject(object::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace container */

namespace core {

template < class FirstType, class SecondType >
struct normalize < container::ObjectsPair < FirstType, SecondType > > {
	static std::pair < object::Object, object::Object > raw2 ( std::pair < FirstType, SecondType > && source ) {
		return std::make_pair ( object::ObjectFactory::make ( std::move ( source.first ) ), object::ObjectFactory::make ( std::move ( source.second ) ) );
	}

	static ext::pair < object::Object, object::Object > raw ( ext::pair < FirstType, SecondType > && source ) {
		return ext::make_pair ( object::ObjectFactory::make ( std::move ( source.first ) ), object::ObjectFactory::make ( std::move ( source.second ) ) );
	}

	static container::ObjectsPair < > eval ( container::ObjectsPair < FirstType, SecondType > && value ) {
		return container::ObjectsPair < > ( raw ( std::move ( value ) ) );
	}
};

} /* namespace core */

#endif /* OBJECTS_PAIR_H_ */
