/*
 * ObjectsTree.h
 *
 * This file is part of Algorithms library toolkit.
 * Copyright (C) 2017 Jan Travnicek (jan.travnicek@fit.cvut.cz)

 * Algorithms library toolkit is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Algorithms library toolkit is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Algorithms library toolkit.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Created on: Apr 1, 2013
 * Author: Jan Travnicek
 */

#ifndef OBJECTS_TREE_H_
#define OBJECTS_TREE_H_

#include <alib/tree>
#include <alib/string>
#include <sstream>

#include "ContainerBase.h"

#include <object/UniqueObject.h>
#include <object/ObjectFactory.h>

#include <core/normalize.hpp>

namespace container {

/**
 * \brief
 * Represents an adaptor of a tree container from the stl extensions provided by the algorithms library.
 *
 * The tree is a structured random access container.
 *
 * \tparam ElementType the type of values in the tree nodes.
 */
template < class ElementType = object::Object >
class ObjectsTree final : public ext::tree < ElementType >, public ContainerBase {
public:
	/**
	 * \brief
	 * Creates a new instance of the tree container wrapper based on the tree datatype from stl extensions of the algorithms library.
	 *
	 * \param raw the extended tree contaier
	 */
	explicit ObjectsTree ( ext::tree < ElementType > raw );

	/**
	 * @copydoc alphabet::ContainerBase::clone ( ) const &
	 */
	virtual ContainerBase * clone ( ) const & override;

	/**
	 * @copydoc alphabet::ContainerBase::clone ( ) &&
	 */
	virtual ContainerBase * clone ( ) && override;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::compare ( const ObjectBase & ) const
	 */
	virtual int compare ( const ObjectBase & other ) const override {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	/**
	 * The actual compare method
	 *
	 * \param other the other instance
	 *
	 * \returns the actual relation between two by type same containers
	 */
	int compare ( const ObjectsTree & other ) const;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::operator >> ( std::ostream & ) const
	 */
	virtual void operator >>( std::ostream & os ) const override;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::operator std::string ( ) const
	 */
	virtual explicit operator std::string ( ) const override;

	/**
	 * @copydoc object::ObjectBase::inc ( ) &&
	 */
	virtual object::ObjectBase * inc ( ) && override;
};

template < class ElementType >
ObjectsTree < ElementType >::ObjectsTree ( ext::tree < ElementType > raw ) : ext::tree < ElementType > ( std::move ( raw ) ) {
}

template < class ElementType >
ContainerBase * ObjectsTree < ElementType >::clone ( ) const & {
	return new ObjectsTree ( * this );
}

template < class ElementType >
ContainerBase * ObjectsTree < ElementType >::clone ( ) && {
	return new ObjectsTree ( std::move ( * this ) );
}

template < class ElementType >
int ObjectsTree < ElementType >::compare ( const ObjectsTree & other ) const {
	static ext::compare < ext::tree < ElementType > > comp;

	return comp ( static_cast < const ext::tree < ElementType > & > ( * this ), static_cast < const ext::tree < ElementType > & > ( other ) );
}

template < class ElementType >
void ObjectsTree < ElementType >::operator >>( std::ostream & os ) const {
	os << "(ObjectsTree " << static_cast < const ext::tree < ElementType > & > ( * this ) << ")";
}

template < class ElementType >
ObjectsTree < ElementType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return std::move ( ss ).str ( );
}

template < class ElementType >
object::ObjectBase* ObjectsTree < ElementType >::inc() && {
	return new object::UniqueObject(object::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace container */

namespace core {

template < class ElementType >
struct normalize < container::ObjectsTree < ElementType > > {
	static ext::tree < object::Object > raw ( ext::tree < ElementType > && source ) {
		ext::vector < ext::tree < object::Object > > children;

		for ( ext::tree < ElementType > & child : source.getChildren ( ) ) {
			children.push_back ( raw ( std::move ( child ) ) );
		}

		return ext::tree < object::Object > ( object::ObjectFactory::make ( std::move ( source.getData ( ) ) ), std::move ( children ) );
	}

	static container::ObjectsTree < > eval ( container::ObjectsTree < ElementType > && value ) {
		return container::ObjectsTree < > ( raw ( std::move ( value ) ) );
	}
};

} /* namespace core */

#endif /* OBJECTS_TREE_H_ */
