/*
 * CommonException.h
 *
 * Created on: Apr 1, 2013
 * Author: Martin Zak
 */

#ifndef COMMON_EXCEPTION_H_
#define COMMON_EXCEPTION_H_

#include <exception>
#include <alib/string>
#include <ostream>

#include "../object/ObjectBase.h"

namespace exception {

/**
 * \brief
 * Basic exception from which all other exceptions are derived.
 *
 * Extends standard exception and additionally contains cause, backtrace (if conputed), and copy of command line string used to execute the program.
 */
class CommonException : public std::exception, public object::ObjectBase {
protected:
	/**
	 * \brief
	 * Programmer supplied identification of the exception cause.
	 */
	std::string m_cause;

	/**
	 * \brief
	 * Overiden string representation of the backtrace from the position of the throw.
	 */
	std::string m_backtrace;

	/**
	 * \brief
	 * A copy of the command line string resulting in the exception.
	 */
	std::string m_command;

	/**
	 * \brief
	 * Accumulation of cause, backtrace, and command into one string.
	 */
	std::string m_whatMessage;

public:
	/**
	 * \brief
	 * Exception constructor with specified cause. The backtrace is internally computed and the command line is also retrieved from global data registry.
	 */
	explicit CommonException ( std::string cause );

	/**
	 * \brief
	 * Constructor allowing specification of all information about the exception, its cause, backtrace of where it happened, and the command executed that resulted in the exception
	 */
	explicit CommonException ( std::string cause, std::string backtrace, std::string command );

	/**
	 * @copydoc object::ObjectBase::clone ( ) const &
	 */
	virtual object::ObjectBase * clone ( ) const & override;

	/**
	 * @copydoc object::ObjectBase::clone ( ) &&
	 */
	virtual object::ObjectBase * clone ( ) && override;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::compare ( const ObjectBase & ) const
	 */
	virtual int compare ( const ObjectBase & other ) const override {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	/**
	 * The actual compare method
	 *
	 * \param other the other instance
	 *
	 * \returns the actual relation between two by type same expression instances
	 */
	int compare ( const CommonException & other ) const;

	/**
	 * \return reason why the exception occurred
	 */
	virtual const char * what ( ) const noexcept override;

	/**
	 * \return reason why the exception occurred
	 */
	const std::string & getCause ( ) const;

	/**
	 * \return reason why the exception occurred
	 */
	const std::string & getBacktrace ( ) const;

	/**
	 * \return reason why the exception occurred
	 */
	const std::string & getCommand ( ) const;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::operator >> ( std::ostream & ) const
	 */
	virtual void operator >>( std::ostream & out ) const override;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::operator std::string ( ) const
	 */
	virtual explicit operator std::string ( ) const override;

	/**
	 * @copydoc object::ObjectBase::inc ( ) &&
	 */
	virtual object::ObjectBase * inc ( ) && override;
};

} /* namespace exception */

#endif /* COMMON_EXCEPTION_H_ */
