/*
 * AnyObject.h
 *
 * This file is part of Algorithms library toolkit.
 * Copyright (C) 2017 Jan Travnicek (jan.travnicek@fit.cvut.cz)

 * Algorithms library toolkit is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Algorithms library toolkit is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Algorithms library toolkit.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Created on: Mar 26, 2013
 *      Author: Jan Travnicek
 */

#ifndef ANY_OBJECT_H_
#define ANY_OBJECT_H_

#include "ObjectBase.h"
#include "Object.h"
#include "UniqueObject.h"

namespace object {

/**
 * \brief
 * Represents an adaptor of any type to a class in type hierarchy of objects in the algorithms library.
 *
 * \tparam T the type of the wrapped object.
 */
template < class T >
class AnyObject final : public ObjectBase {
	/**
	 * \brief
	 * The wrapped object.
	 */
	T m_data;
public:
	/**
	 * \brief
	 * Constructor of the class based on the value of the wrapped object.
	 *
	 * \param data the object to be wrapped
	 */
	explicit AnyObject ( const T & data );

	/**
	 * \brief
	 * Constructor of the class based on the value of the wrapped object.
	 *
	 * \param data the object to be wrapped
	 */
	explicit AnyObject ( T && data );

	/**
	 * @copydoc ObjectBase::clone ( ) const &
	 */
	virtual ObjectBase * clone ( ) const & override;

	/**
	 * @copydoc ObjectBase::clone ( ) &&
	 */
	virtual ObjectBase * clone ( ) && override;

	/**
	 * Getter of the wrapped object
	 *
	 * \return reference to the wrapped object.
	 */
	const T & getData ( ) const;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::compare ( const ObjectBase & ) const
	 */
	virtual int compare ( const ObjectBase & other ) const override {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	/**
	 * The actual compare method
	 *
	 * \param other the other instance
	 *
	 * \returns the actual relation between two by type same containers
	 */
	int compare ( const AnyObject & other ) const;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::operator >> ( std::ostream & ) const
	 */
	virtual void operator >>( std::ostream & out ) const override;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::operator std::string ( ) const
	 */
	virtual explicit operator std::string ( ) const override;

	/**
	 * @copydoc object::ObjectBase::inc ( ) &&
	 */
	virtual ObjectBase * inc ( ) && override;
};

template < class T >
AnyObject < T >::AnyObject ( const T & data ) : m_data ( data ) {
}

template < class T >
AnyObject < T >::AnyObject ( T && data ) : m_data ( std::move ( data ) ) {
}

template < class T >
ObjectBase* AnyObject < T >::clone ( ) const & {
	return new AnyObject(*this);
}

template < class T >
ObjectBase* AnyObject < T >::clone() && {
	return new AnyObject(std::move(*this));
}

template < class T >
const T & AnyObject < T >::getData ( ) const {
	return m_data;
}

template < class T >
int AnyObject < T >::compare(const AnyObject < T > & other ) const {
	static ext::compare < decltype ( m_data ) > comp;

	return comp ( m_data, other.m_data );
}

template < class T >
void AnyObject < T >::operator>>(std::ostream& out) const {
	out << "(AnyObject" << m_data << ")";
}

template < class T >
AnyObject < T > ::operator std::string () const {
	return ext::to_string ( m_data );
}

template < class T >
ObjectBase* AnyObject < T >::inc() && {
	return new UniqueObject(Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace object */

#endif /* ANY_OBJECT_H_ */
