/*
 * UniqueObject.h
 *
 * This file is part of Algorithms library toolkit.
 * Copyright (C) 2017 Jan Travnicek (jan.travnicek@fit.cvut.cz)

 * Algorithms library toolkit is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Algorithms library toolkit is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Algorithms library toolkit.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Created on: Nov 11, 2016
 *      Author: Jan Travnicek
 */

#ifndef UNIQUE_OBJECT_H_
#define UNIQUE_OBJECT_H_

#include "Object.h"
#include <primitive/Integer.h>

namespace object {

/**
 * \brief
 * Represents object based on other object with additional differentiation counter.
 *
 * The UniqueObject is used in situations when some object is given and one would like to create new based on it but different. For example q vs. q' vs q'', etc.
 *
 * Primarly, the class is used from inc method.
 */
class UniqueObject final : public ObjectBase {
	/**
	 * The base object.
	 */
	Object m_object;

	/**
	 * The differentiating number.
	 */
	primitive::Integer m_id;

public:
	/**
	 * \brief
	 * Creates new object based on one provided by the \p object and the differentiating number.
	 *
	 * \param object the base object
	 * \param id the differentiating number
	 */
	explicit UniqueObject ( Object object, primitive::Integer id );

	/**
	 * @copydoc object::ObjectBase::clone ( ) const &
	 */
	virtual ObjectBase * clone ( ) const & override;

	/**
	 * @copydoc object::ObjectBase::clone ( ) &&
	 */
	virtual ObjectBase * clone ( ) && override;

	/**
	 * Getter of the base object.
	 *
	 * \return the base object
	 */
	const Object & getObject ( ) const &;

	/**
	 * Getter of the base object.
	 *
	 * \return the base object
	 */
	Object && getObject ( ) &&;

	/**
	 * Getter of the differentiating number.
	 *
	 * \return the differentiating number
	 */
	const primitive::Integer & getId ( ) const;

	/**
	 * Getter of the differentiating number.
	 *
	 * \return the differentiating number
	 */
	primitive::Integer & getId ( );

	/**
	 * @copydoc base::CommonBase < ObjectBase >::compare ( const ObjectBase & ) const
	 */
	virtual int compare ( const ObjectBase & other ) const override {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	/**
	 * The actual compare method
	 *
	 * \param other the other instance
	 *
	 * \returns the actual relation between two by type same objects
	 */
	int compare ( const UniqueObject & other ) const;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::operator >> ( std::ostream & ) const
	 */
	virtual void operator >>( std::ostream & ) const override;

	/**
	 * @copydoc base::CommonBase < ObjectBase >::operator std::string ( ) const
	 */
	virtual explicit operator std::string ( ) const override;

	/**
	 * @copydoc object::ObjectBase::inc ( ) &&
	 */
	virtual ObjectBase * inc ( ) && override;
};

} /* namespace object */

#endif /* UNIQUE_OBJECT_H_ */
