/*
 * deque.hpp
 *
 * This file is part of Algorithms library toolkit.
 * Copyright (C) 2017 Jan Travnicek (jan.travnicek@fit.cvut.cz)

 * Algorithms library toolkit is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Algorithms library toolkit is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Algorithms library toolkit.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Created on: Apr 1, 2013
 * Author: Jan Travnicek
 */

#ifndef __DEQUE_HPP_
#define __DEQUE_HPP_

#include <deque>
#include <ostream>
#include <sstream>
#include <string>

#include "compare.hpp"
#include "allocFix.hpp"

namespace ext {

/**
 * Class extending the deque class from the standard library. Original reason is to allow printing of the container with overloaded operator <<.
 *
 * The class mimics the behavior of the deque from the standatd library.
 *
 * \tparam T the type of values inside the deque
 * \tparam Ts ... remaining unimportant template parameters of the deque
 */
template < class T, class Alloc = std::allocator < T > >
class deque : public std::deque < T, Alloc >, AllocFix < Alloc > {
public:
	/**
	 * Inherit constructors of the standard deque
	 */
	using std::deque < T, Alloc >::deque;

	/**
	 * Inherit operator = of the standard deque
	 */
	using std::deque < T, Alloc >::operator =;
#ifndef __clang__

	/**
	 * Default constructor needed by g++ since it is not inherited
	 */
	deque ( ) = default;

	/**
	 * Copy constructor needed by g++ since it is not inherited
	 */
	deque ( const deque & other ) = default;

	/**
	 * Move constructor needed by g++ since it is not inherited
	 */
	deque ( deque && other ) = default;

	/**
	 * Copy operator = needed by g++ since it is not inherited
	 */
	deque & operator = ( deque && other ) = default;

	/**
	 * Move operator = needed by g++ since it is not inherited
	 */
	deque & operator = ( const deque & other ) = default;
#endif
};

/**
 * Operator to print the deque to the output stream.
 *
 * \param out the output stream
 * \param deque the deque to print
 *
 * \tparam T the type of values inside the deque
 * \tparam Ts ... remaining unimportant template parameters of the deque
 *
 * \return the output stream from the \p out
 */
template < class T, class ... Ts >
std::ostream & operator <<( std::ostream & out, const ext::deque < T, Ts ... > & list ) {
	out << "[";

	bool first = true;

	for ( const T & item : list ) {
		if ( !first ) out << ", ";

		first = false;
		out << item;
	}

	out << "]";
	return out;
}

/**
 * Specialisation of the compare structure implementing the three-way comparison
 *
 * \tparam T the type of values inside the deque
 * \tparam Ts ... remaining unimportant template parameters of the deque
 */
template < class T, class ... Ts >
struct compare < ext::deque < T, Ts ... > > {

	/**
	 * Implementation of the three-way comparison
	 *
	 * \param first the left operand of the comparison
	 * \param second the right operand of the comparison
	 *
	 * \return negative value of left < right, positive value if left > right, zero if left == right
	 */
	int operator ()( const ext::deque < T, Ts ... > & first, const ext::deque < T, Ts ... > & second ) const {
		if ( first.size ( ) < second.size ( ) ) return -1;

		if ( first.size ( ) > second.size ( ) ) return 1;

		static compare < typename std::decay < T >::type > comp;

		for ( auto iterF = first.begin ( ), iterS = second.begin ( ); iterF != first.end ( ); ++iterF, ++iterS ) {
			int res = comp ( * iterF, * iterS );

			if ( res != 0 ) return res;
		}

		return 0;
	}

};

/**
 * Overload of to_string function.
 *
 * \param value the deque to be converted to string
 *
 * \tparam T the type of values inside the deque
 * \tparam Ts ... remaining unimportant template parameters of the deque
 */
template < class T, class ... Ts >
std::string to_string ( const ext::deque < T, Ts ... > & value ) {
	std::stringstream ss;
	ss << value;
	return ss.str();
}

} /* namespace ext */

#endif /* __DEQUE_HPP_ */
