/*
 * tree.hpp
 *
 * Created on: Jul 2, 2016
 * Author: Jan Travnicek
 */

#ifndef __TREE_BASE_HPP_
#define __TREE_BASE_HPP_

#include "ptr_vector.hpp"
#include "ptr_array.hpp"

namespace ext {

template < class Data >
class BaseNode {
	Data * parent;

	template < class D, class C >
	friend class NullaryNode;
	template < class D, class C >
	friend class UnaryNode;
	template < class D, class C >
	friend class BinaryNode;
	template < class D, class C >
	friend class TernaryNode;
	template < class D, std::size_t I, class C >
	friend class AnyaryNode;
	template < class D, class C >
	friend class FixedaryNode;
	template < class D, class C >
	friend class VararyNode;

	Data * operator ->( ) {
		return static_cast < Data * > ( this );
	}

	const Data * operator ->( ) const {
		return static_cast < const Data * > ( this );
	}

public:
	BaseNode ( ) : parent ( nullptr ) {
	}

	virtual ~BaseNode ( ) noexcept {
	}

	BaseNode ( const BaseNode & ) : parent ( nullptr ) {
	}

	BaseNode ( BaseNode && ) noexcept : parent ( nullptr ) {
	}

	BaseNode & operator =( const BaseNode & ) {
		return * this;
	}

	BaseNode & operator =( BaseNode && ) noexcept {
		return * this;
	}

	Data * getParent ( ) {
		return parent;
	}

	const Data * getParent ( ) const {
		return parent;
	}

};

template < class Data, std::size_t arity, class Cast = Data >
class AnyaryNode {
	ext::ptr_array < Data, arity > m_children;

	void setParents ( ) {
		for ( Data & child : m_children )
			child->parent = static_cast < Cast * > ( this );
	}

public:
	AnyaryNode ( ext::ptr_array < Data, arity > c ) : m_children ( std::move ( c ) ) {
		setParents ( );
	}

	virtual ~AnyaryNode ( ) noexcept {
	}

	AnyaryNode ( const AnyaryNode & other ) : AnyaryNode ( other.m_children ) {
	}

	AnyaryNode ( AnyaryNode && other ) noexcept : AnyaryNode ( std::move ( other.m_children ) ) {
	}

	AnyaryNode & operator =( const AnyaryNode & other ) {
		return * this = AnyaryNode ( other );
	}

	AnyaryNode & operator =( AnyaryNode && other ) noexcept {
		using std::swap;

		swap ( this->m_children, other.m_children );

		setParents ( );

		return * this;
	}

	const ext::ptr_array < Data, arity > & getElements ( ) {
		return m_children;
	}

	const ext::ptr_array < Data, arity > & getElements ( ) const {
		return m_children;
	}

	template < int N >
	const Data & getElement ( ) const {
		return std::get < N > ( m_children );
	}

	template < int N >
	Data & getElement ( ) {
		return std::get < N > ( m_children );
	}

	void setElements ( ext::ptr_array < Data, arity > c ) {
		m_children = std::move ( c );
		setParents ( );
	}

	template < int N >
	void setElement ( const Data & d ) {
		m_children.set ( m_children.cbegin ( ) + N, d );
		std::get < N > ( m_children )->parent = static_cast < Cast * > ( this );
	}

	template < int N >
	void setElement ( Data && d ) {
		m_children.set ( m_children.cbegin ( ) + N, std::move ( d ) );
		std::get < N > ( m_children )->parent = static_cast < Cast * > ( this );
	}

};

template < class Data, class Cast = Data >
class NullaryNode : public AnyaryNode < Data, 0, Cast > {
public:
	NullaryNode ( ) : AnyaryNode < Data, 0, Cast > ( ext::make_ptr_array < Data > ( ) ) {
	}

};

template < class Data, class Cast = Data >
class UnaryNode : public AnyaryNode < Data, 1, Cast > {
public:
	UnaryNode ( const Data & c ) : AnyaryNode < Data, 1, Cast > ( ext::make_ptr_array ( c ) ) {
	}

	UnaryNode ( Data && c ) : AnyaryNode < Data, 1, Cast > ( ext::make_ptr_array ( std::move ( c ) ) ) {
	}

	Data & getChild ( ) {
		return this->template getElement < 0 > ( );
	}

	const Data & getChild ( ) const {
		return this->template getElement < 0 > ( );
	}

	void setChild ( const Data & c ) {
		this->template setElement < 0 > ( c );
	}

	void setChild ( Data && c ) {
		this->template setElement < 0 > ( std::move ( c ) );
	}

};

template < class Data, class Cast = Data >
class BinaryNode : public AnyaryNode < Data, 2, Cast > {
public:
	BinaryNode ( const Data & l, const Data & r ) : AnyaryNode < Data, 2, Cast > ( ext::make_ptr_array ( l, r ) ) {
	}

	BinaryNode ( Data && l, Data && r ) : AnyaryNode < Data, 2, Cast > ( ext::make_ptr_array ( std::move ( l ), std::move ( r ) ) ) {
	}

	Data & getLeft ( ) {
		return this->template getElement < 0 > ( );
	}

	const Data & getLeft ( ) const {
		return this->template getElement < 0 > ( );
	}

	void setLeft ( const Data & l ) {
		this->template setElement < 0 > ( l );
	}

	void setLeft ( Data && l ) {
		this->template setElement < 0 > ( std::move ( l ) );
	}

	Data & getRight ( ) {
		return this->template getElement < 1 > ( );
	}

	const Data & getRight ( ) const {
		return this->template getElement < 1 > ( );
	}

	void setRight ( const Data & r ) {
		this->template setElement < 1 > ( r );
	}

	void setRight ( Data && r ) {
		this->template setElement < 1 > ( std::move ( r ) );
	}

};

template < class Data, class Cast = Data >
class TernaryNode : public AnyaryNode < Data, 3, Cast > {
public:
	TernaryNode ( const Data & f, const Data & s, const Data & t ) : AnyaryNode < Data, 3, Cast > ( ext::make_ptr_array ( f, s, t ) ) {
	}

	TernaryNode ( Data && f, Data && s, Data && t ) : AnyaryNode < Data, 3, Cast > ( ext::make_ptr_array ( std::move ( f ), std::move ( s ), std::move ( t ) ) ) {
	}

	Data & getFirst ( ) {
		return this->template getElement < 0 > ( );
	}

	const Data & getFirst ( ) const {
		return this->template getElement < 0 > ( );
	}

	void setFirst ( const Data & f ) {
		this->template setElement < 0 > ( f );
	}

	void setFirst ( Data && f ) {
		this->template setElement < 0 > ( std::move ( f ) );
	}

	Data & getSecond ( ) {
		return this->template getElement < 1 > ( );
	}

	const Data & getSecond ( ) const {
		return this->template getElement < 1 > ( );
	}

	void setSecond ( const Data & s ) {
		this->template setElement < 1 > ( s );
	}

	void setSecond ( Data && s ) {
		this->template setElement < 1 > ( std::move ( s ) );
	}

	Data & getThird ( ) {
		return this->template getElement < 2 > ( );
	}

	const Data & getThird ( ) const {
		return this->template getElement < 2 > ( );
	}

	void setThird ( const Data & t ) {
		this->template setElement < 2 > ( t );
	}

	void setThird ( Data && t ) {
		this->template setElement < 2 > ( std::move ( t ) );
	}

};

template < class Data, class Cast = Data >
class FixedaryNode {
	ext::ptr_vector < Data > m_children;

public:
	template < typename ... Types >
	FixedaryNode ( Types ... data ) {
		Data args[] { std::move ( data ) ... };

		for ( Data & child : args )
			m_children.push_back ( std::move ( child ) );

		for ( Data & child : m_children )
			child.parent = static_cast < Cast * > ( this );
	}

	FixedaryNode ( ext::ptr_vector < Data > c ) : m_children ( std::move ( c ) ) {
		for ( Data & child : m_children )
			child.parent = static_cast < Cast * > ( this );
	}

	virtual ~FixedaryNode ( ) noexcept {
	}

	FixedaryNode ( const FixedaryNode & other ) : FixedaryNode ( other.m_children ) {
	}

	FixedaryNode ( FixedaryNode && other ) noexcept : FixedaryNode ( std::move ( other.m_children ) ) {
	}

	FixedaryNode & operator =( const FixedaryNode & other ) {
		return * this = FixedaryNode ( other );
	}

	FixedaryNode & operator =( FixedaryNode && other ) noexcept {
		using std::swap;

		swap ( this->m_children, other.m_children );

		for ( Data & child : m_children )
			child.parent = static_cast < Cast * > ( this );

		return * this;
	}

	ext::ptr_vector < Data > & getChildren ( ) {
		return m_children;
	}

	const ext::ptr_vector < Data > & getChildren ( ) const {
		return m_children;
	}

	void setChildren ( ext::ptr_vector < Data > c ) {
		if ( c.size ( ) != m_children.size ( ) )
			throw "Arity != size";

		m_children = std::move ( c );

		for ( Data & child : m_children )
			child->parent = static_cast < Cast * > ( this );
	}

	void setChild ( Data && d, typename ext::ptr_vector < Data >::const_iterator it ) {
		m_children.set ( it, std::move ( d ) )->parent = static_cast < Cast * > ( this );
	}

	void setChild ( const Data & d, typename ext::ptr_vector < Data >::const_iterator it ) {
		m_children.set ( it, d )->parent = static_cast < Cast * > ( this );
	}

	void setChild ( const Data & d, int index ) {
		setChild ( d, m_children.begin ( ) + index );
	}

	void setChild ( Data && d, int index ) {
		setChild ( std::move ( d ), m_children.begin ( ) + index );
	}

};

template < class Data, class Cast = Data >
class VararyNode {
	ext::ptr_vector < Data > m_children;

public:
	VararyNode ( ) {
	}

	VararyNode ( ext::ptr_vector < Data > c ) : m_children ( std::move ( c ) ) {
		for ( Data & child : m_children )
			child.parent = static_cast < Cast * > ( this );
	}

	virtual ~VararyNode ( ) noexcept {
	}

	VararyNode ( const VararyNode & other ) : VararyNode ( other.m_children ) {
	}

	VararyNode ( VararyNode && other ) noexcept : VararyNode ( std::move ( other.m_children ) ) {
	}

	VararyNode & operator =( const VararyNode & other ) {
		return * this = VararyNode ( other );
	}

	VararyNode & operator =( VararyNode && other ) noexcept {
		using std::swap;

		swap ( this->m_children, other.m_children );

		for ( Data & child : m_children )
			child.parent = static_cast < Cast * > ( this );

		return * this;
	}

	ext::ptr_vector < Data > & getChildren ( ) {
		return m_children;
	}

	const ext::ptr_vector < Data > & getChildren ( ) const {
		return m_children;
	}

	void setChildren ( ext::ptr_vector < Data > c ) {
		m_children = std::move ( c );

		for ( Data & child : m_children )
			child.parent = static_cast < Cast * > ( this );
	}

	void setChild ( Data && d, typename ext::ptr_vector < Data >::const_iterator it ) {
		m_children.set ( it, std::move ( d ) )->parent = static_cast < Cast * > ( this );
	}

	void setChild ( const Data & d, typename ext::ptr_vector < Data >::const_iterator it ) {
		m_children.set ( it, d )->parent = static_cast < Cast * > ( this );
	}

	void setChild ( const Data & d, int index ) {
		setChild ( d, m_children.begin ( ) + index );
	}

	void setChild ( Data && d, int index ) {
		setChild ( std::move ( d ), m_children.begin ( ) + index );
	}

	typename ext::ptr_vector < Data >::iterator insert ( typename ext::ptr_vector < Data >::const_iterator it, const Data & d ) {
		typename ext::ptr_vector < Data >::iterator iter = m_children.insert ( it, d );

		iter->parent = static_cast < Cast * > ( this );

		return iter;
	}

	typename ext::ptr_vector < Data >::iterator insert ( typename ext::ptr_vector < Data >::const_iterator it, Data && d ) {
		typename ext::ptr_vector < Data >::iterator iter = m_children.insert ( it, std::move ( d ) );

		iter->parent = static_cast < Cast * > ( this );

		return iter;
	}

	template < class InputIterator >
	typename ext::ptr_vector < Data >::iterator insert ( typename ext::ptr_vector < Data >::const_iterator it, InputIterator first, InputIterator last ) {
		size_t size = std::distance ( first, last );

		typename ext::ptr_vector < Data >::iterator iter = m_children.insert ( it, first, last );

		for ( size_t i = 0; i < size; ++ i ) {
			( iter + i )->parent = static_cast < Cast * > ( this );
		}

		return iter;
	}

	void pushBackChild ( const Data & d ) {
		m_children.push_back ( d );
		m_children [ m_children.size ( ) - 1].parent = static_cast < Cast * > ( this );
	}

	void pushBackChild ( Data && d ) {
		m_children.push_back ( std::move ( d ) );
		m_children [ m_children.size ( ) - 1].parent = static_cast < Cast * > ( this );
	}

};

template < class Data, class Cast >
class Tree {
public:
	static void setChild ( const UnaryNode < Data, Cast > & node, Data child ) {
		const_cast < UnaryNode < Data, Cast > & > ( node ).setChild ( std::move ( child ) );
	}

	static void setLeft ( const BinaryNode < Data, Cast > & node, Data child ) {
		const_cast < BinaryNode < Data, Cast > & > ( node ).setLeft ( std::move ( child ) );
	}

	static void setRight ( const BinaryNode < Data, Cast > & node, Data child ) {
		const_cast < BinaryNode < Data, Cast > & > ( node ).setRight ( std::move ( child ) );
	}

	static void setFirst ( const TernaryNode < Data, Cast > & node, Data child ) {
		const_cast < TernaryNode < Data, Cast > & > ( node ).setFirst ( std::move ( child ) );
	}

	static void setSecond ( const TernaryNode < Data, Cast > & node, Data child ) {
		const_cast < TernaryNode < Data, Cast > & > ( node ).setSecond ( std::move ( child ) );
	}

	static void setThird ( const TernaryNode < Data, Cast > & node, Data child ) {
		const_cast < TernaryNode < Data, Cast > & > ( node ).setThird ( std::move ( child ) );
	}

	template < std::size_t arity >
	static void setChildren ( const AnyaryNode < Data, arity, Cast > & node, ext::ptr_array < Data, arity > children ) {
		const_cast < AnyaryNode < Data, arity, Cast > & > ( node ).setChildren ( std::move ( children ) );
	}

	template < class N, std::size_t arity >
	static void setChild ( const AnyaryNode < Data, arity, Cast > & node, Data child ) {
		const_cast < AnyaryNode < Data, arity, Cast > & > ( node ).template setChild < N > ( std::move ( child ) );
	}

	static void setChildren ( const FixedaryNode < Data, Cast > & node, ext::ptr_vector < Data > child ) {
		const_cast < FixedaryNode < Data, Cast > & > ( node ).setChildren ( std::move ( child ) );
	}

	static void setChild ( const FixedaryNode < Data, Cast > & node, Data && child, int index ) {
		const_cast < FixedaryNode < Data, Cast > & > ( node ).setChild ( std::move ( child ), index );
	}

	static void setChildren ( const VararyNode < Data, Cast > & node, ext::ptr_vector < Data > children ) {
		const_cast < VararyNode < Data, Cast > & > ( node ).setChildren ( std::move ( children ) );
	}

	static void setChild ( const VararyNode < Data, Cast > & node, Data && child, int index ) {
		const_cast < VararyNode < Data, Cast > & > ( node ).setChild ( std::move ( child ), index );
	}

	static void setChild ( const VararyNode < Data, Cast > & node, Data && child, typename ext::ptr_vector < Data >::iterator it ) {
		const_cast < VararyNode < Data, Cast > & > ( node ).setChild ( std::move ( child ), it );
	}

	typename ext::ptr_vector < Data >::iterator insert ( const VararyNode < Data, Cast > & node, typename ext::ptr_vector < Data >::iterator it, Data && child ) {
		const_cast < VararyNode < Data, Cast > & > ( node ).insert ( it, std::move ( child ) );
	}

	template < class InputIterator >
	typename ext::ptr_vector < Data >::iterator insert ( const VararyNode < Data, Cast > & node, typename ext::ptr_vector < Data >::iterator it, InputIterator first, InputIterator last ) {
		const_cast < VararyNode < Data, Cast > & > ( node ).insert ( it, first, last );
	}

	static void pushBackChild ( const VararyNode < Data, Cast > & node, Data && child ) {
		const_cast < VararyNode < Data, Cast > & > ( node ).pushBackChild ( std::move ( child ) );
	}

};

} /* namespace ext */

#endif /* __TREE_BASE_HPP_ */
