/*
 * XmlApi.hpp
 *
 * Created on: Apr 1, 2013
 * Author: Jan Travnicek
 */

#ifndef XML_API_HPP_
#define XML_API_HPP_

#include <alib/functional>
#include <alib/deque>
#include <alib/map>
#include <alib/string>
#include <alib/memory>
#include "sax/Token.h"
#include "sax/FromXMLParserHelper.h"

#include "base/CommonBase.hpp"
#include "base/WrapperBase.hpp"

#include <global/GlobalData.h>

#include "exception/CommonException.h"

namespace core {

template < typename T, typename Enable = void >
struct xmlApi { };

class xmlApiInputContext : public ext::deque < sax::Token >::iterator {
	template < typename T, typename Enable >
	friend struct xmlApi;

	ext::map < std::string, ext::map < int, base::WrapperBaseBase * > > idToInstanceContexts;
	int idMaxContext;

public:
	xmlApiInputContext ( ext::deque < sax::Token >::iterator iter ) : ext::deque < sax::Token >::iterator ( iter ), idMaxContext ( 0 ) {
	}

	~xmlApiInputContext ( ) {
		for ( const std::pair < const std::string, ext::map < int, base::WrapperBaseBase * > > context : idToInstanceContexts )
			for ( const std::pair < const int, base::WrapperBaseBase * > entry : context.second )
				delete entry.second;
	}

	ext::map < int, base::WrapperBaseBase * > & idToInstance ( const std::string & name ) {
		return idToInstanceContexts[name];
	}

	int & idMax ( ) {
		return idMaxContext;
	}

	const ext::map < std::string, ext::map < int, base::WrapperBaseBase * > > & dump ( ) const {
		return idToInstanceContexts;
	}

};

class xmlApiOutputContext : public ext::deque < sax::Token > {
	template < typename T, typename Enable >
	friend struct xmlApi;

	ext::map < std::string, ext::map < const base::CommonBaseBase *, int > > instanceToIdContexts;
	int idMaxContext;

public:
	xmlApiOutputContext ( ) : idMaxContext ( 0 ) {
	}

	ext::map < const base::CommonBaseBase *, int > & instanceToId ( const std::string & name ) {
		return instanceToIdContexts[name];
	}

	int & idMax ( ) {
		return idMaxContext;
	}

};

template < typename Group >
struct xmlApi < Group, typename std::enable_if < std::is_base_of < base::WrapperBaseBase, Group >::value >::type > {
private:
	class GroupParser {
	public:
		virtual Group parse ( ext::deque < sax::Token >::iterator & input ) = 0;

		virtual ~GroupParser ( ) {
		}
	};

	static ext::map < std::string, std::unique_ptr < GroupParser > > & parseFunctions ( ) {
		static ext::map < std::string, std::unique_ptr < GroupParser > > res;

		return res;
	}

	template < class Type >
	class ParserRegister : public GroupParser {
		std::function < Type ( ext::deque < sax::Token >::iterator & ) > parseFunction;

	public:
		ParserRegister( ) : parseFunction ( xmlApi < Type >::parse ) {
		}

		virtual ~ParserRegister( ) {
		}

		virtual Group parse ( ext::deque < sax::Token >::iterator & input ) {
			return Group ( parseFunction ( input ) );
		}

	};

	class GroupComposer {
	public:
		virtual void compose ( ext::deque < sax::Token > & input, const Group & group ) = 0;

		virtual ~GroupComposer( ) {
		}
	};

	static ext::map < std::string, std::unique_ptr < GroupComposer > > & composeFunctions ( ) {
		static ext::map < std::string, std::unique_ptr < GroupComposer > > res;

		return res;
	}

	template < class Type >
	class ComposerRegister : public GroupComposer {
		std::function < void ( ext::deque < sax::Token > &, const Type & ) > composeFunction;

	public:
		ComposerRegister( ) : composeFunction ( xmlApi < Type >::compose ) {
		}

		virtual ~ComposerRegister ( ) {
		}

		virtual void compose ( ext::deque < sax::Token > & input, const Group & group ) {
			composeFunction ( input, static_cast < const Type & > ( group.getData ( ) ) );
		}

	};

public:
	template < class Type >
	static void registerXmlReader ( ) {
		bool res = parseFunctions ( ).insert ( std::make_pair ( xmlApi < Type >::xmlTagName(), std::unique_ptr < GroupParser > ( new ParserRegister < Type > ( ) ) ) ).second;
		if ( ! res ) {
			std::string groupName = ext::to_string < Group > ( );
			std::string typeName = ext::to_string < Type > ( );

			throw::exception::CommonException ( "Parse callback of " + typeName + " already registered in group " + groupName + "." );
		}
	}

	template < class Type >
	static void registerXmlWriter ( ) {
		bool res = composeFunctions ( ).insert ( std::make_pair ( ext::to_string < Type > ( ), std::unique_ptr < GroupComposer > ( new ComposerRegister < Type > ( ) ) ) ).second;
		if ( ! res ) {
			std::string groupName = ext::to_string < Group > ( );
			std::string typeName = ext::to_string < Type > ( );

			throw::exception::CommonException ( "Parse callback of " + typeName + " already registered in group " + groupName + "." );
		}
	}

	static Group parse ( ext::deque < sax::Token >::iterator & data ) {
		xmlApiInputContext & input = ( xmlApiInputContext & ) data;
		sax::FromXMLParserHelper::skipAttributes ( input, sax::Token::TokenType::START_ELEMENT );

		const std::string & tagName = sax::FromXMLParserHelper::getTokenData ( input, sax::Token::TokenType::START_ELEMENT );
		if ( /* common::GlobalData::optimizeXml && */ tagName == "Ref" ) {
			sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "Ref" );
			sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ATTRIBUTE, "id" );
			int id = ext::from_string < int > ( sax::FromXMLParserHelper::popTokenData ( input, sax::Token::TokenType::CHARACTER ) );
			sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ATTRIBUTE, "id" );
			sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "Ref" );
			ext::map < int, base::WrapperBaseBase * >::iterator elem = input.idToInstance ( ext::to_string < Group > ( ) ).find ( id );

			if ( elem == input.idToInstance ( ext::to_string < Group > ( ) ).end ( ) ) {
				common::Streams::err << input.dump ( ) << std::endl;
				throw exception::CommonException ( "XML Inconsistent ( id not found " + ext::to_string ( id  ) + " )" );
			}

			return * ( ( Group * ) elem->second );
		} else {
			typename ext::map < std::string, std::unique_ptr < GroupParser > >::iterator callback = parseFunctions ( ).find ( tagName );

			if ( callback == parseFunctions ( ).end ( ) ) throw exception::CommonException ( "Parse callback for " + tagName + " tag not registered." );

			/* find out if this is a base for reference */
			bool ref = sax::FromXMLParserHelper::isToken ( input + 1, sax::Token::TokenType::START_ATTRIBUTE, "ref" );
			int id = 0;
			if ( ref ) {
				ext::deque < sax::Token >::iterator tmp = input + 2;
				id = ext::from_string < int > ( sax::FromXMLParserHelper::popTokenData ( tmp, sax::Token::TokenType::CHARACTER ) );
			}

			Group res ( callback->second->parse ( input ) );

			/* if object is a base of reference, register it */
			if ( ref )
				input.idToInstance ( ext::to_string < Group > ( ) ).insert ( std::make_pair ( id, ( base::WrapperBaseBase * ) new Group ( res ) ) );

			return res;
		}
	}

	static bool first ( const ext::deque < sax::Token >::const_iterator & input ) {
		const std::string & tagName = sax::FromXMLParserHelper::getTokenData ( input, sax::Token::TokenType::START_ELEMENT );

		if ( tagName == "Ref" ) return true;

		typename ext::map < std::string, std::unique_ptr < GroupParser > >::iterator callback = parseFunctions ( ).find ( tagName );

		if ( callback == parseFunctions ( ).end ( ) ) return false;

		return true;
	}

	static std::string xmlTagName ( ) {
		std::string target = ext::to_string < Group > ( );

		throw ::exception::CommonException ( "Type " + target + " does not have xmlTagName." );
	}

	static void compose ( ext::deque < sax::Token > & output, const Group & data ) {
		xmlApiOutputContext & context = ( xmlApiOutputContext & ) output;

		typename ext::map < const base::CommonBaseBase *, int >::iterator elem = context.instanceToId ( ext::to_string < Group > ( ) ).find ( static_cast < const base::CommonBaseBase * > ( & data.getData ( ) ) );

		if ( common::GlobalData::optimizeXml && elem != context.instanceToId ( ext::to_string < Group > ( ) ).end ( ) ) {
			output.emplace_back ( "Ref", sax::Token::TokenType::START_ELEMENT );
			output.emplace_back ( "id", sax::Token::TokenType::START_ATTRIBUTE );
			output.emplace_back ( ext::to_string ( elem->second ), sax::Token::TokenType::CHARACTER );
			output.emplace_back ( "id", sax::Token::TokenType::END_ATTRIBUTE );
			output.emplace_back ( "Ref", sax::Token::TokenType::END_ELEMENT );
		} else {
			/* get current position in token stream */
			size_t pos = output.size();
			int id = 0;
			if ( common::GlobalData::optimizeXml ) {
				id = context.idMax ( )++;
			}

			std::string type = ext::to_string ( ext::type_index ( typeid ( data.getData ( ) ) ) );
			typename ext::map < std::string, std::unique_ptr < GroupComposer > >::iterator callback = composeFunctions ( ).find ( type );

			if ( callback == composeFunctions ( ).end ( ) ) throw exception::CommonException ( "Compose callback for " + type + " tag not registered." );

			/* encode referenced object */
			callback->second->compose ( output, data );

			/* emplace referenced attribute */
			if ( common::GlobalData::optimizeXml ) {
				output.emplace ( output.begin ( ) + pos + 1, "ref", sax::Token::TokenType::START_ATTRIBUTE );
				output.emplace ( output.begin ( ) + pos + 2, ext::to_string ( id ), sax::Token::TokenType::CHARACTER );
				output.emplace ( output.begin ( ) + pos + 3, "ref", sax::Token::TokenType::END_ATTRIBUTE );

				context.instanceToId ( ext::to_string < Group > ( ) ).insert ( std::make_pair ( static_cast< const base::CommonBaseBase * > ( & data.getData ( ) ), id ) );
			}
		}
	}

};

} /* namespace core */

#endif /* XML_API_HPP_ */
