/*
 * aquery.cpp
 *
 *  Created on: 26. 3. 2014
 *	  Author: Jan Travnicek
 */

#include <tclap/CmdLine.h>
#include <global/GlobalData.h>
#include <measure>

#include <exception/CommonException.h>
#include <lexer/Lexer.h>
#include <parser/Parser.h>

#include <factory/XmlDataFactory.hpp>

int main ( int argc, char * argv[] ) {
	try {
		common::GlobalData::argc = argc;
		common::GlobalData::argv = argv;

		TCLAP::CmdLine cmd ( "Stringology algorithm access binary", ' ', "0.01" );

		std::vector < std::string > allowed;
		allowed.push_back ( "suffixTrieFactors" );
		allowed.push_back ( "suffixArrayFactors" );
		allowed.push_back ( "suffixAutomatonFactors" );
		allowed.push_back ( "positionHeapFactors" );
		allowed.push_back ( "bitParallelismFactors" );
		allowed.push_back ( "compressedBitParallelismFactors" );
		allowed.push_back ( "compressedBitParallelismPatterns" );
		allowed.push_back ( "nonlinearCompressedBitParallelismPatterns" );
		allowed.push_back ( "fullAndLinearIndexPatterns" );
		allowed.push_back ( "nonlinearFullAndLinearIndexPatterns" );
		TCLAP::ValuesConstraint < std::string > allowedVals ( allowed );

		TCLAP::ValueArg < std::string > query ( "q", "query", "Query index", false, "exactFactorMatch", & allowedVals );
		cmd.add ( query );

		TCLAP::ValueArg <std::string > indexInput ( "i", "index", "Index to query", false, "-", "file");
		cmd.add( indexInput );

		TCLAP::MultiArg < std::string > patternInput ( "p", "pattern", "Pattern object from file", false, "file" );
		cmd.add ( patternInput );

		TCLAP::SwitchArg measure ( "m", "measure", "Measure times", false );
		cmd.add ( measure );

		TCLAP::SwitchArg verbose ( "v", "verbose", "Be verbose", false );
		cmd.add ( verbose );

		cmd.parse ( argc, argv );

		if(verbose.isSet())
			common::GlobalData::verbose = true;
		if(measure.isSet())
			common::GlobalData::measure = true;

		cli::Environment environment;
		environment.setBinding ( "indexData", indexInput.getValue ( ) );
		environment.setBinding ( "stdout", "-" );

		measurements::start ( "Overal", measurements::Type::OVERALL );
		measurements::start ( "Input read", measurements::Type::AUXILIARY );

		cli::Parser parser ( cli::Lexer ( "execute <#indexData > $index" ) );
		parser.parse ( )->run ( environment );

		if ( query.getValue ( ) == "suffixTrieFactors"
		  || query.getValue ( ) == "suffixArrayFactors"
		  || query.getValue ( ) == "suffixAutomatonFactors"
		  || query.getValue ( ) == "positionHeapFactors"
		  || query.getValue ( ) == "bitParallelismFactors"
		  || query.getValue ( ) == "compressedBitParallelismFactors"
		  || query.getValue ( ) == "compressedBitParallelismPatterns"
		  || query.getValue ( ) == "nonlinearCompressedBitParallelismPatterns"
		  || query.getValue ( ) == "fullAndLinearIndexPatterns"
		  || query.getValue ( ) == "nonlinearFullAndLinearIndexPatterns" ) {
			std::string input;
			if ( patternInput.getValue ( ).size ( ) == 0 )
				input = "-";
			else if ( patternInput.getValue ( ).size ( ) == 1 )
				input = * patternInput.getValue ( ).begin ( );
			else
				throw exception::CommonException("Multiple parameters when single required.");

			environment.setBinding ( "patternData", input );
			parser = cli::Parser ( cli::Lexer ( "execute <#patternData > $pattern" ) );
			parser.parse ( )->run ( environment );
		}

		measurements::end ( );
		measurements::start ( "Algorithm", measurements::Type::MAIN );

		std::string cliCommand;
		if ( query.getValue ( ) == "suffixTrieFactors" ) {
			cliCommand = "execute stringology::query::SuffixTrieFactors $index $pattern > $output";
		} else if ( query.getValue ( ) == "suffixArrayFactors" ) {
			cliCommand = "execute stringology::query::SuffixArrayFactors $index $pattern > $output";
		} else if ( query.getValue ( ) == "suffixAutomatonFactors" ) {
			cliCommand = "execute stringology::query::SuffixAutomatonFactors $index $pattern > $output";
		} else if ( query.getValue ( ) == "positionHeapFactors" ) {
			cliCommand = "execute stringology::query::PositionHeapFactors $index $pattern > $output";
		} else if ( query.getValue ( ) == "bitParallelismFactors" ) {
			cliCommand = "execute stringology::query::BitParallelismFactors $index $pattern > $output";
		} else if ( query.getValue ( ) == "compressedBitParallelismFactors" ) {
			cliCommand = "execute stringology::query::CompressedBitParallelismFactors $index $pattern > $output";
		} else if ( query.getValue ( ) == "compressedBitParallelismPatterns" ) {
			cliCommand = "execute arbology::query::CompressedBitParallelismPatterns $index $pattern > $output";
		} else if ( query.getValue ( ) == "nonlinearCompressedBitParallelismPatterns" ) {
			cliCommand = "execute arbology::query::NonlinearCompressedBitParallelismPatterns $index $pattern > $output";
		} else if ( query.getValue ( ) == "fullAndLinearIndexPatterns" ) {
			cliCommand = "execute arbology::query::FullAndLinearIndexPatterns $index $pattern > $output";
		} else if ( query.getValue ( ) == "nonlinearFullAndLinearIndexPatterns" ) {
			cliCommand = "execute arbology::query::NonlinearFullAndLinearIndexPatterns $index $pattern > $output";
		} else {
			throw exception::CommonException ( "Invalid algorithm" );
		}

		parser = cli::Parser ( cli::Lexer ( cliCommand ) );
		parser.parse ( )->run ( environment );

		measurements::end ( );
		measurements::start ( "Output write", measurements::Type::AUXILIARY );

		parser = cli::Parser ( cli::Lexer ( "execute $output >#stdout" ) );
		parser.parse ( )->run ( environment );

		measurements::end ( );
		measurements::end ( );

		if ( measure.getValue ( ) ) common::Streams::measure << measurements::results ( ) << std::endl;

		return 0;
	} catch ( const exception::CommonException & exception ) {
		factory::XmlDataFactory::toStdout ( exception );
		return 1;
	} catch ( const TCLAP::ArgException & exception ) {
		common::Streams::out << exception.error ( ) << std::endl;
		return 2;
	} catch ( const std::exception & exception ) {
		common::Streams::err << "Exception caught: " << exception.what ( ) << std::endl;
		return 3;
	} catch ( ... ) {
		common::Streams::err << "Unknown exception caught." << std::endl;
		return 127;
	}
}
